// HMMLook.cpp: implementation of the CHMMLook class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "qlook.h"
#include "HMMLook.h"
#include <math.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CHMMLook::CHMMLook()
{
	m_nCurModel=0;
	m_nTestedValues=256; // quantity of tested values
	m_NextHMMLook = NULL;
}

CHMMLook::~CHMMLook()
{

}


/* ------------------------ Lexical Scanner ------------------------ */

#define MAXSYMLEN 40

/* Internal and binary keyword representation */
typedef enum {   /* Only a character big !! */
   BEGINHMM, USEMAC, ENDHMM, NUMMIXES, 
   NUMSTATES, STREAMINFO, VECSIZE, 
   NDUR, PDUR, GDUR, RELDUR, GENDUR,
   DIAGCOV,  FULLCOV, XFORMCOV,
   STATE, TMIX, MIXTURE, STREAM, SWEIGHTS,
   MEAN, VARIANCE, INVCOVAR, X_FORM, GCONST,
   DURATION, INVDIAGCOV, TRANSP, DPROB, LLTCOV, LLTCOVAR,
   XFORMKIND=90, PARENTXFORM, NUMXFORMS, XFORMSET,
   LINXFORM, OFFSET, BIAS, BLOCKINFO, BLOCK, BASECLASS, 
   CLASS, XFORMWGTSET, CLASSXFORM, MMFIDMASK, PARAMETERS,
   NUMCLASSES, ADAPTKIND, PREQUAL, INPUTXFORM,
   RCLASS=110, REGTREE, NODE, TNODE,
   HMMSETID=119,
   PARMKIND=120, 
   MACRO, EOFSYM, NULLSYM   /* Special Syms - not literals */
} Symbol;

/* Mapping between verbose keywords for readable external HMM format */
/*  and internal/external binary token numbers */
static struct {
   char *name;
   Symbol sym; } symMap[] = 
   { { "BEGINHMM", BEGINHMM }, { "USE", USEMAC }, 
     { "ENDHMM", ENDHMM }, { "NUMMIXES", NUMMIXES }, 
     { "NUMSTATES", NUMSTATES }, { "STREAMINFO", STREAMINFO }, 
     { "VECSIZE", VECSIZE }, { "NULLD", NDUR }, 
     { "POISSOND", PDUR }, { "GAMMAD", GDUR }, 
     { "RELD", RELDUR }, { "GEND", GENDUR }, 
     { "DIAGC", DIAGCOV }, {  "FULLC", FULLCOV }, 
     { "XFORMC", XFORMCOV }, { "STATE", STATE }, 
     { "TMIX", TMIX }, { "MIXTURE", MIXTURE }, 
     { "STREAM", STREAM }, { "SWEIGHTS", SWEIGHTS }, 
     { "MEAN", MEAN }, { "VARIANCE", VARIANCE }, 
     { "INVCOVAR", INVCOVAR }, { "XFORM", X_FORM }, 
     { "GCONST" , GCONST }, { "DURATION", DURATION }, 
     { "INVDIAGC", INVDIAGCOV }, { "TRANSP", TRANSP }, 
     { "DPROB", DPROB }, { "LLTC", LLTCOV }, 
     { "LLTCOVAR", LLTCOVAR }, 
     { "RCLASS", RCLASS }, { "REGTREE", REGTREE }, 
     { "NODE", NODE }, { "TNODE", TNODE }, 
     { "HMMSETID", HMMSETID }, 
     { "PARMKIND", PARMKIND }, { "MACRO", MACRO }, 
     { "EOF", EOFSYM }, 
     /* Transformation symbols */
     {"XFORMKIND", XFORMKIND } , {"PARENTXFORM", PARENTXFORM },
     {"NUMXFORMS", NUMXFORMS }, {"XFORMSET", XFORMSET },
     {"LINXFORM", LINXFORM }, {"OFFSET", OFFSET }, 
     {"BIAS", BIAS }, {"BLOCKINFO", BLOCKINFO }, {"BLOCK", BLOCK }, 
     {"BASECLASS", BASECLASS }, {"CLASS", CLASS }, 
     {"XFORMWGTSET", XFORMWGTSET }, {"CLASSXFORM", CLASSXFORM }, 
     {"MMFIDMASK", MMFIDMASK }, {"PARAMETERS", PARAMETERS },
     {"NUMCLASSES", NUMCLASSES }, {"ADAPTKIND", ADAPTKIND },
     {"PREQUAL", PREQUAL }, {"INPUTXFORM", INPUTXFORM },
     { "", NULLSYM } };
#define NUMSYM (sizeof(symMap)/sizeof(symMap[0]))

CHMMModelSet::CHMMModelSet(){
	m_nChannels=0;
}

// auxiliary functions
bool ReadInt(int& nValue, char* buf, int& nPos)
{
	int n, i,j, sign=0;
	char c;
	for(i=n=0; c = buf[nPos+i]; i++){
		if(c >= 0 && c <= 32){// whitespace
			if(sign !=0){break;}// something was read!
			continue; // scan to 1st non white symbol
		}
		if(c== '-'){
			if(sign !=0){break;}// next number started
			sign = -1; continue;
		}
		if(c== '+'){
			if(sign !=0){break;}// next number started
			sign = 1; continue;
		}
		if(c >= 0x030 && c <= 0x039){
			j = c - 0x030;
			if(!j){// do not increase grades for leading zeroes
				if(sign == 0){continue;}
			}else{
				if(!sign){sign = 1;}
			}
			n = n*10 + j;
		}else{
			if(sign){break;}
		}
	}// next char
	if(!sign){return false;}
	nValue = n * sign;
	nPos+=i;
	return true;
}
bool ReadFloat(float& fValue, char* buf, int& nPos)
{
	int n, i,j, sign=0, nfrac=0, nexp=0, signexp=0;
	float frac=0.;
	bool bexp=false;
	char c;
	for(i=n=0; c = buf[nPos+i]; i++){
		if(c >= 0 && c <= 32){// whitespace
			if(sign !=0){break;}// something was read!
			continue; // scan to 1st non white symbol
		}
		if(c== '-'){
			if(bexp){
				if(!signexp){
					signexp = -1; continue;
				}else{// next number started
					break;
				}
			}
			if(sign !=0){break;}// next number started
			sign = -1; continue;
		}
		if(c== '+'){
			if(bexp){
				if(!signexp){
					signexp = 1; continue;
				}else{// next number started
					break;
				}
			}
			if(sign !=0){break;}// next number started
			sign = 1; continue;
		}
		if(c== '.' || c== ','){
			if(frac !=0){break;}// next number started
			nfrac = 1;  continue;
		}
		if(c== 'e' || c== 'E'){// degree
			if(sign ==0){break;}// next number started
			bexp = true; continue;
		}
		if(c >= 0x030 && c <= 0x039){
			j = c - 0x030;
			if(bexp){
				nexp = nexp * 10 + j;continue;
			}
			if(!j){// do not increase grades for leading zeroes
				if(sign == 0){continue;}
			}else{
				if(!sign){sign = 1;}
			}
			if(nfrac){
				frac = frac*10 + j;
				nfrac *= 10;
			}else{
				n = n*10 + j;
			}
		}else{
			if(sign){break;}
		}
	}// next char
	if(!sign){return false;}
	if(frac){frac /= nfrac;}
	fValue = (n + frac) * sign ;
	if(signexp){
		for(j=0; j<nexp; j++){
			if(signexp > 0){fValue *= 10;
			}else{fValue /= 10;
			}
		}
	}
	nPos+=i;
	return true;
}
int ReadFloatVector(float* pf, int nMaxSize, char* buf, int& nPos){
	int i;
	for(i=0; i<nMaxSize; i++){
		if(!ReadFloat(pf[i], buf, nPos)){// not float data happened
			break;
		}
	}
	return i;
}

Symbol ReadTag(char* buf, int& nPos)
{
	int i,j;
	bool bInTag=false;
	char c;
//	static 
	char Tag[32];
	Symbol sym = NULLSYM;
	Tag[0]='\0';
	for(i=j=0; c=buf[nPos + i]; i++){
		if(c>=0 && c<= 32){continue;}// skip white char
		if(!bInTag){
			if(c != '<'){break;}
			bInTag = true;
			continue;
		}
		if(c != '>'){
			ASSERT(j < 32);
			Tag[j++] = c;
			continue;
		}
		// last tag name
		Tag[j] = '\0';
		// find a tag in symMap
		for(j = 0; j < NUMSYM; j++){
			if(strcmp(symMap[j].name, Tag) == 0){
				sym = symMap[j].sym;
				break;
			}
		}
		break;
	}
	if(sym != NULLSYM){
		nPos+=i;
	}
	return sym;
}

int GetQuotedString(char* sCurHMM, const char* buf, int nPos){
	ASSERT(sCurHMM);
	int i,j;
	char c;
	int nInString=0;
	for(i=j=0; c=buf[nPos + i]; i++){
		if(c >= 0 && c <= 32){continue;}// skip white char
		if(c == '"'){
			if(nInString){
				sCurHMM[j] = '\0';
				break;
			}else{
				nInString++;
				continue;
			}
		}
		if(nInString){
			sCurHMM[j++] = c;
		}
	}
	nPos+=i+1;
	return j;
}
/////////////////////////////


int CHMMState::AddMix(int nDim, const float* pMean, const float* pVariance)
{
	int n = m_Mixes.GetSize();
	m_Mixes.SetSize(n + 1); 
	CMix& mix = m_Mixes[n];
	mix.m_Mean.SetSize(nDim);
	if(pMean){memcpy( mix.m_Mean.GetData(), pMean, nDim*sizeof(pMean[0]) );}
	mix.m_Mean.SetSize(nDim);
	if(pVariance){memcpy( mix.m_Variance.GetData(), pVariance, nDim*sizeof(pVariance[0]) );}
	return n;
}

enum {// load errors
	err_load_no,
	err_load_generic,
	err_load_hierarchy_broken,
	err_load_extra_state,
	err_load_no_mix_weight,
	err_load_channel_differs,

	err_load_last
};

/////////////////////////////////////////////
bool CHMMModelSet::CopyState(CHMMModel* pCurHmm, int nState, char* sCurState)
{
	int i, nMix;
	// find the common state by name
	for(i=0; i<m_CommonStateNames.GetSize(); i++){
		if(m_CommonStateNames[i] == sCurState){break;}
	}
	if(i >= m_CommonStateNames.GetSize()){ASSERT(false); return false;}
	CHMMState& StateSrc = m_CommonStates[i]; // source
	CHMMState* pStateTgt = pCurHmm->GetState(nState); // target
	CMix* pMixSrc;
	CMix* pMixTgt;
	if(!pStateTgt){ASSERT(false); return false;}
	for(nMix = 0; nMix < StateSrc.m_Mixes.GetSize(); nMix++){
		pMixSrc = StateSrc.GetMix(nMix);
		pStateTgt->AddMix(pMixSrc->m_Mean.GetSize(), 0, 0);
		VERIFY(pMixTgt = pStateTgt->GetMix(nMix));
		pMixTgt->SetVariance(pMixSrc->m_Variance.GetSize(),0);
		if(!pMixSrc){return false;}
		for(i = 0; i < pMixSrc->m_Mean.GetSize(); i++){
			pMixTgt->m_Mean[i] = pMixSrc->m_Mean[i];
			pMixTgt->m_Variance[i] = pMixSrc->m_Variance[i];
		}
	}
	return true;
}
/////////////////////////////////////////
//
int CHMMModelSet::Load(LPCTSTR sPath, DWORD dwFlags)
{
//	CArray<CHMMModel, CHMMModel&>;
//	CArray<CFloatVector, CFloatVector&> m_Mean;
	int nret = -1;

	FILE* pf = fopen(sPath, "r");
	if(!pf){return nret;}

	enum {maxbuflen = 1024*5};
	char buf[maxbuflen]; // buffer to read to
	bool bOnlyWorkingStates = true; // TODO: depending on dwFlags
	int nPos; // current position inside the buf
	int nChannels=0, nCurChannels;
	int nCurHMM=0, nStates, nCurState=-1, nCurCommonState=-1, nCurMix, nMixes;
	CHMMState* pCurState;
	CMix* pCurMix;
	float* means=0; float * vars=0; float weight;
	char sCurHMM[32]; sCurHMM[0] = '\0';
	char sCurState[32];sCurState[0] = '\0';
	Symbol CurSymbol = NULLSYM;
	char cMacroType = '\0';
	bool bNeedLine=true;
	CHMMModel* pCurHmm=0;
	int nTransp;

	while(true)
	{// read a line from hmm macro file
		if(bNeedLine){
			if(!fgets(buf, maxbuflen, pf)){
				ASSERT(!cMacroType);
				break;
			}
			nPos = 0;
			bNeedLine=true;
		}

		switch(cMacroType){

		case 's': // read a common state
			if(buf[nPos]  == '~'){
				cMacroType = '\0'; sCurState[0] = '\0'; bNeedLine=false; break;
			}
			if(!sCurState[0]){// read hmm name
				GetQuotedString(sCurState, buf, nPos);
				if(nCurState >= 0){// we are inside a model. just substitute the state
					CopyState(pCurHmm, nCurState, sCurState);
					cMacroType = 'h'; bNeedLine=true;
					sCurState[0] = '\0';
					break;
				}
				nCurCommonState = m_CommonStates.GetSize();
				m_CommonStates.SetSize(nCurCommonState + 1);
				m_CommonStateNames.Add(sCurState);
				CurSymbol = NULLSYM; pCurMix = 0;
				bNeedLine = true;
				pCurState = &m_CommonStates[nCurCommonState];
				nCurChannels = 0; pCurMix=0;
				nCurState=0; nCurMix=0; nMixes=1;
			}
			if(CurSymbol == NULLSYM){
				CurSymbol = ReadTag(buf, nPos);
			}
			switch(CurSymbol){ // process data according to the tag type
			case NUMMIXES:
				if(!ReadInt(nMixes, buf, nPos)){nret =-err_load_generic; goto load_error;}
				CurSymbol = NULLSYM;
				break;

			case MIXTURE:
//				if(!pCurHmm){nret =-err_load_hierarchy_broken; goto load_error;}
//				pCurState = pCurHmm->GetState(nCurState);
				if(!ReadInt(nCurMix, buf, nPos)){nret =-err_load_generic; goto load_error;}
				nCurMix -= 1;
				nCurMix = pCurState->AddMix(nChannels, 0, 0);
				pCurMix = pCurState->GetMix(nCurMix);
				if(!pCurMix){nret =-err_load_generic; goto load_error;}
				if(!ReadFloat(weight, buf, nPos)){
					nret =-err_load_no_mix_weight; goto load_error;
				}
				pCurMix->m_weight = weight;
				CurSymbol = NULLSYM;
				break;

			case MEAN:
				if(!nCurChannels){
					if(!ReadInt(nCurChannels, buf, nPos)){nret =-err_load_generic; goto load_error;}
					if(nChannels != nCurChannels){
						nret =-err_load_channel_differs; goto load_error;
					}
					break;
				}
				if(!pCurMix){
					nCurMix = pCurState->AddMix(nChannels, 0, 0);
					pCurMix = pCurState->GetMix(nCurMix);
				}
				if(!means){means = new float[nChannels];}
				ReadFloatVector(means, nCurChannels, buf, nPos);
				// update mix with mean
				pCurMix->SetMean(nChannels, means);
				CurSymbol = NULLSYM; nCurChannels = 0;
				break;

			case VARIANCE:
				if(!nCurChannels){
					if(!ReadInt(nCurChannels, buf, nPos)){nret =-err_load_generic; goto load_error;}
					if(nChannels != nCurChannels){
						nret =-err_load_channel_differs; goto load_error;
					}
					break;
				}
				if(!pCurMix){nret =-err_load_generic; goto load_error;}
				if(!vars){vars = new float[nChannels];}
				ReadFloatVector(vars, nCurChannels, buf, nPos);
				// update mix with variances
				pCurMix->SetVariance(nChannels, vars);
				CurSymbol = NULLSYM; nCurChannels = 0;
//				sCurState[0] = '\0';
				break;

			case GCONST:
				if(!pCurMix){nret =-err_load_generic; goto load_error;}
				if(!ReadFloat(pCurMix->m_gconst, buf, nPos)){
					nret =-err_load_generic; goto load_error;}
				CurSymbol = NULLSYM;
//				sCurState[0] = '\0';
				break;

			default :
				bNeedLine = true;
				break;
			}// END symbols for 's'

			break; // END of case 's'

		///////////////////////////////////////////////////////////////
		case 'h': // read a model
			if(buf[nPos]  == '~'){
				cMacroType = '\0'; sCurState[0] = '\0'; bNeedLine=false; break;
			}
			if(!sCurHMM[0]){// read hmm name
				GetQuotedString(sCurHMM, buf, nPos);
				bNeedLine = true;
			}
			if(CurSymbol == NULLSYM){
				CurSymbol = ReadTag(buf, nPos);
			}

			switch(CurSymbol){ // process data according to the tag type

			case BEGINHMM:
				nCurHMM++; nCurChannels = 0; pCurMix=0;
				pCurState=0; nCurState=0; nCurMix=0; nMixes=1;
				// insert a new hmm
				pCurHmm = &m_Models[AddModel()];
				pCurHmm->SetName(sCurHMM);
				CurSymbol = NULLSYM;
				break;

			case ENDHMM:
				pCurHmm = 0; sCurHMM[0] = '\0';
				nCurState = -1; pCurState = 0; pCurMix = 0;
				CurSymbol = NULLSYM;
				cMacroType = '\0';
				break;

			case NUMSTATES:
				if(!ReadInt(nStates, buf, nPos)){nret =-err_load_generic; goto load_error;}
				if(bOnlyWorkingStates){nStates -= 2;}
				nStates = pCurHmm->SetStateCount(nStates);
				CurSymbol = NULLSYM;
				break;

			case STATE:
				if(!ReadInt(nCurState, buf, nPos)){nret =-err_load_generic; goto load_error;}
				if(!bOnlyWorkingStates){// TODO: add 0-time passing states
					nCurState-=1;
					ASSERT(false);
				}else{
					nCurState -= 2;
				}
				pCurState = pCurHmm->GetState(nCurState);
				CurSymbol = NULLSYM; pCurMix = 0;
				break;

			case NUMMIXES:
				if(!ReadInt(nMixes, buf, nPos)){nret =-err_load_generic; goto load_error;}
				CurSymbol = NULLSYM;
				break;

			case MIXTURE:
				if(!pCurHmm){nret =-err_load_hierarchy_broken; goto load_error;}
				pCurState = pCurHmm->GetState(nCurState);
				if(!ReadInt(nCurMix, buf, nPos)){nret =-err_load_generic; goto load_error;}
				nCurMix -= 1;
				nCurMix = pCurState->AddMix(nChannels, 0, 0);
				pCurMix = pCurState->GetMix(nCurMix);
				if(!pCurMix){nret =-err_load_generic; goto load_error;}
				if(!ReadFloat(weight, buf, nPos)){
					nret =-err_load_no_mix_weight; goto load_error;
				}
				pCurMix->m_weight = weight;
				CurSymbol = NULLSYM;
				break;

			case MEAN:
				if(!nCurChannels){
					if(!ReadInt(nCurChannels, buf, nPos)){nret =-err_load_generic; goto load_error;}
					if(nChannels != nCurChannels){
						nret =-err_load_channel_differs; goto load_error;
					}
					break;
				}
				if(!pCurMix){
					nCurMix = pCurState->AddMix(nChannels, 0, 0);
					pCurMix = pCurState->GetMix(nCurMix);
				}
				if(!means){means = new float[nChannels];}
				ReadFloatVector(means, nCurChannels, buf, nPos);
				// update mix with mean
				pCurMix->SetMean(nChannels, means);
				CurSymbol = NULLSYM; nCurChannels = 0;
				break;

			case VARIANCE:
				if(!nCurChannels){
					if(!ReadInt(nCurChannels, buf, nPos)){nret =-err_load_generic; goto load_error;}
					if(nChannels != nCurChannels){
						nret =-err_load_channel_differs; goto load_error;
					}
					break;
				}
				if(!pCurMix){nret =-err_load_generic; goto load_error;}
				if(!vars){vars = new float[nChannels];}
				ReadFloatVector(vars, nCurChannels, buf, nPos);
				// update mix with variances
				pCurMix->SetVariance(nChannels, vars);
				CurSymbol = NULLSYM; nCurChannels = 0;
				break;

			case GCONST:
				if(!pCurMix){nret =-err_load_generic; goto load_error;}
				if(!ReadFloat(pCurMix->m_gconst, buf, nPos)){
					nret =-err_load_generic; goto load_error;}
				CurSymbol = NULLSYM;
				break;

			case TRANSP:
				if(!pCurHmm){nret =-err_load_hierarchy_broken; goto load_error;}
				if(!ReadInt(nTransp, buf, nPos)){
					nret =-err_load_generic; goto load_error;}
				{
					pCurHmm->SetTranspSize(nTransp);
					int i;
					for(i=0; i<nTransp; i++){
						// get next line
						if(!fgets(buf, maxbuflen, pf)){
							nret =-err_load_generic; goto load_error;
						}
						ASSERT(nTransp < nChannels && vars);
						ReadFloatVector(vars, nTransp, buf, nPos);
						pCurHmm->SetTransp(i, vars);
					}
				}
				bNeedLine = true;
				CurSymbol = NULLSYM;
				break;

			default :
				bNeedLine = true;
				break;

			}// switch(CurSymbol) (tags)
			break; // 'h' macro

		case 'o':
			if(!bNeedLine){bNeedLine=true; continue;}
//			break;
			if(CurSymbol == NULLSYM){
				CurSymbol = ReadTag(buf, nPos);
			}
			switch(CurSymbol){ // process data according to the tag type

			case STREAMINFO:
				CurSymbol = NULLSYM;
				break;

			case VECSIZE:
				if(!ReadInt(nChannels, buf, nPos)){nret =-err_load_generic; goto load_error;}
				m_nChannels = nChannels;
				CurSymbol = NULLSYM;
				cMacroType = '\0';
				break;

			case NDUR:
				break;

			case NULLSYM:
				bNeedLine = true;
				break;

			default :
				break;
			}
			break;

		case '\0':
//			sCurHMM[0]='\0';
			CurSymbol = NULLSYM;
			if(buf[0] != '~'){break;}
			cMacroType = buf[1];
			nPos = 2;
			bNeedLine = false;
			break;
		default:
			ASSERT(false);
			break;
		}// switch(cMacroType)
	}// next line
	goto load_clear;

load_error:
//	nret = -1;
load_clear:
	if(means) {delete []means;}
	if(vars) {delete []vars;}
	return nret;
}


int CHMMLook::LoadHMMs(LPCTSTR sPath, DWORD dwFlags)
{
	int nret = m_HmmSet.Load(sPath);
	if(nret == 0){m_nCurModel = 0;}
	return nret;
}

// fill look structures accordingly to loaded hmms
int CHMMLook::DoAnalyseHmms()
{
	int nret = -1;
	
	FillVectorsMinMax();
	// for energy channel, live its original min-max (TODO: customize)
//	int nExceptChannel = 12; 
//	FillVectorsMinMaxGlobally(nExceptChannel);
	UnifyMinMaxForChannels(0, 12);
	UnifyMinMaxForChannels(13, 13);
	UnifyMinMaxForChannels(26, 13);

	AnalyseModel();

	return nret;
}

int CHMMLook::FillVectorsMinMaxGlobally(int nExceptChannel){
	//CFloatVector m_arrValueMinMax;
	int nChan;
	float mean, variance, min, max, value;
	min = 1000000, max = -min;
	for(nChan=0; nChan<m_HmmSet.m_nChannels; nChan++)
	{
		if(nChan == nExceptChannel){continue;}
		value = m_arrValueMinMax[nChan * 2];
		if(value < min){	
			min = value;
		}
		value = m_arrValueMinMax[nChan * 2 + 1];
		if(value > max){
			max = value;
		}
	}
	TRACE("Global min = %f; max = %f\n", min, max);
	for(nChan=0; nChan<m_HmmSet.m_nChannels; nChan++)
	{
		if(nChan == nExceptChannel){continue;}
		m_arrValueMinMax[nChan * 2] = min;
		m_arrValueMinMax[nChan * 2 + 1] = max;
	}
	return 0;
}
int CHMMLook::UnifyMinMaxForChannels(int nStartChannel, int nChannels)
{
	//CFloatVector m_arrValueMinMax;
	int nChan, i;
	float mean, variance, min, max, value;
	if(nChannels == 0){return 0;}
	if(m_arrValueMinMax.GetSize() < m_HmmSet.m_nChannels * 2){
		ASSERT(false);
		return -1;
	}
	if(nChannels < 0){nChannels = m_HmmSet.m_nChannels - nStartChannel;}
	min = 1000000, max = -min;
	for(i=0; i<nChannels; i++)
	{
		nChan = nStartChannel + i;
		if(nChan >= m_HmmSet.m_nChannels){break;}
		value = m_arrValueMinMax[nChan * 2];
		if(value < min){	
			min = value;
		}
		value = m_arrValueMinMax[nChan * 2 + 1];
		if(value > max){
			max = value;
		}
	}
	TRACE("For channels %d to %d: min = %f; max = %f\n", 
		nStartChannel, nStartChannel + nChannels - 1, min, max);
	for(i=0; i<nChannels; i++)
	{
		nChan = nStartChannel + i;
		if(nChan >= m_HmmSet.m_nChannels){break;}
		m_arrValueMinMax[nChan * 2] = min;
		m_arrValueMinMax[nChan * 2 + 1] = max;
	}
	return 0;
}
int CHMMLook::FillVectorsMinMax(){
	//CFloatVector m_arrValueMinMax; 
	int nChan, nModel, nState, nMixture;
	float mean, variance, min, max, value;
	CHMMModel* pCurHmm;
	CHMMState* pCurState;
	CMix* pCurMix;
	float variance_emphasis = 3;
	m_arrValueMinMax.SetSize(m_HmmSet.m_nChannels * 2);
	for(nChan=0; nChan<m_HmmSet.m_nChannels; nChan++)
	{// for each channel
		min = 1000000, max = -min;
		for(nModel=0; nModel<m_HmmSet.m_Models.GetSize(); nModel++){
			pCurHmm = &m_HmmSet.m_Models[nModel];
			for(nState=0; nState<pCurHmm->m_States.GetSize(); nState++){
				pCurState = &pCurHmm->m_States[nState];
				for(nMixture=0; pCurMix = pCurState->GetMix(nMixture); nMixture++){
					mean = pCurMix->m_Mean[nChan];
					variance = pCurMix->m_Variance[nChan];
					value = mean + sqrt(variance) * variance_emphasis;
					if(value > max){
						max = value;
					}
					value = mean - sqrt(variance) * variance_emphasis;
					if(value < min){	
						min = value;
					}
				}// NEXT nMixture
			}// NEXT nState
		}// NEXT nModel
		m_arrValueMinMax[nChan * 2] = min;
		m_arrValueMinMax[nChan * 2 + 1] = max;
	}// NEXT nChan
	TRACE(_T("Values min-max:\n"));
	for(nChan=0; nChan<m_HmmSet.m_nChannels; nChan++){
		TRACE(_T("%d: ( %f : %f )\n"), nChan, 
			m_arrValueMinMax[nChan * 2], m_arrValueMinMax[nChan * 2 + 1]);
	}
	return 0;
}

int GetClosestIdxTo(float value, float* arr, int size){
	if(!arr || size<=0){return -1;}
	int nret = 0;
	float dmin = value - arr[0];
	float d;
	if(dmin < 0){dmin *= -1 ;};
	for(int i=1; i<size; i++){
		d = value - arr[i];
		if(d < 0){d *= -1;}
		if(d < dmin){dmin = d; nret = i;}
	}
	return nret;
}

int CHMMLook::AnalyseModel()
{
	if(m_nCurModel<0 || m_nCurModel >= m_HmmSet.m_Models.GetSize()){return -1;}
	CHMMModel& model = m_HmmSet.m_Models[m_nCurModel];
	int nStates = model.m_States.GetSize();
	int nChannels = m_HmmSet.m_nChannels;
	int i,j, nState, nChan, nMix, nMixes;
	float value_min, value_max;
	CFloatVector arrf;

	// allocate m_TestedStates
	m_TestedStates.SetSize(nStates);
	for(i=0; i<nStates; i++){
		m_TestedStates[i].SetSize(m_nTestedValues * nChannels);
	}
	m_arrStateMeanIdxes.SetSize(nStates);
	m_arrChannelValueZeroIdxes.SetSize(nStates);
	// fill data
	for(nState=0; nState<nStates; nState++){
//		float* pout = m_TestedStates[nState].GetData();
		CHMMState& state = model.m_States[nState];
		CStateMeanIdxes& ChannelMeansIdxes = m_arrStateMeanIdxes[nState];
		nMixes = state.m_Mixes.GetSize();
		ChannelMeansIdxes.m_nMixes = nMixes;
		ChannelMeansIdxes.m_ChannelMeansIdxes.SetSize(nChannels * nMixes);
		m_arrChannelValueZeroIdxes[nState].SetSize(nChannels);
		for(nChan=0; nChan<nChannels; nChan++)
		{
			value_min = m_arrValueMinMax[nChan * 2]; 
			value_max = m_arrValueMinMax[nChan * 2 + 1];
			float b = // koef for linear value running between min and max 
				( value_max - value_min ) / (m_nTestedValues - 1); 
			
			// get indexes of tested values closest to zero and to means
			arrf.SetSize(m_nTestedValues);
			for(j=0; j<m_nTestedValues; j++){
				arrf[j] = value_min + b * j;
			}
			j = GetClosestIdxTo(0, arrf.GetData(), m_nTestedValues); // zero
			arrf[j] = arrf[j];
			m_arrChannelValueZeroIdxes[nState][nChan] = j;
			for(nMix = 0; nMix < nMixes; nMix++){
				CMix& mix = *(state.GetMix(nMix));
				j = GetClosestIdxTo(mix.m_Mean[nChan], arrf.GetData(), m_nTestedValues); // zero
				arrf[j] = arrf[j];
				ChannelMeansIdxes.m_ChannelMeansIdxes[nChan*nMixes + nMix] = j;
			} // NEXT mix

			for(j=0; j<m_nTestedValues; j++){
				float fprob = 0;
				float value = value_min + b * j;
				for(nMix = 0; nMix < state.m_Mixes.GetSize(); nMix++){
					CMix& mix = *(state.GetMix(nMix));
					fprob += mix.m_weight * GetValueProb(
						value, mix.m_Mean[nChan], mix.m_Variance[nChan]);
				} // NEXT mix
				m_TestedStates[nState][m_nTestedValues * nChan + j] = fprob;
//				pout[m_nTestedValues * nChan + j] = fout;
			}// NEXT value
		}// NEXT channel
	}// next state
	return 0;
}

float* CHMMLook::GetProbValues(int nState, int& nValues){
	if(nState<0 || nState>=m_TestedStates.GetSize()){return 0;}
	nValues = m_nTestedValues;
	return m_TestedStates[nState].GetData();
}

float CHMMLook::GetValueProb(float x, float mean, float sigma)
{
#define PI 3.14159
	//PI = acos(-1);
	float fret = (1/sqrt(2*PI*sigma)) * exp(-.5 * (x-mean) * (x-mean) / sigma);
	return fret;
}
