// HMMLook.h: interface for the CHMMLook class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_HMMLOOK_H__D33E6AAB_01D6_463C_B928_93DBF2C31ACD__INCLUDED_)
#define AFX_HMMLOOK_H__D33E6AAB_01D6_463C_B928_93DBF2C31ACD__INCLUDED_

#include <afxtempl.h>

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

typedef CArray<float, float&> CFloatVector;

class CMix
{
public:
	CMix(){m_gconst=0; m_weight=1;};
	virtual ~CMix(){};

	void SetMean(int nDim, const float* pMean){
		m_Mean.SetSize(nDim);
		if(pMean){memcpy( m_Mean.GetData(), pMean, nDim*sizeof(pMean[0]) );}
	};

	void SetVariance(int nDim, const float* pVariance){
		m_Variance.SetSize(nDim);
		if(pVariance){memcpy( m_Variance.GetData(), pVariance, nDim*sizeof(pVariance[0]) );}
	}

	float m_weight;
	CArray<float, float&> m_Mean;
	CArray<float, float&> m_Variance;
	float m_gconst;
};

class CHMMState
{
public:
	CHMMState(){};
	virtual ~CHMMState(){};

	int AddMix(int nDim, const float* pMean, const float* pVariance);
	CMix* GetMix(int n)
	{return (n>=0 && n<m_Mixes.GetSize()) ? &m_Mixes[n] : 0;};

	CArray<CMix, CMix&> m_Mixes;
};

class CHMMModel
{
public:
	CHMMModel(){};
	virtual ~CHMMModel(){};
	int SetStateCount(int nStates){
		ASSERT(nStates>0);m_States.SetSize(nStates);
		return nStates;
	};

	int GetStateCount(){return m_States.GetSize();};
	CHMMState* GetState(int n)
	{return (n>=0 && n<m_States.GetSize()) ? &m_States[n] : 0;};
	void SetName(LPCTSTR sName){m_sName = sName;};
	LPCTSTR GetName(){return LPCTSTR(m_sName);};
	void SetTranspSize(int nTransp){
		m_Transp.SetSize(nTransp);
		for(int i=0; i<nTransp; i++){
			m_Transp[i].SetSize(nTransp);
		}
	}
	void SetTransp(int nTransp, const float* pTransp){
		m_Transp.SetSize(m_Transp.GetSize()); // square matrix
		if(pTransp){
			memcpy( m_Transp[nTransp].GetData(), pTransp, m_Transp[nTransp].GetSize()*sizeof(pTransp[0]) );
		}
	}

	CArray<CHMMState, CHMMState&> m_States;
	CArray<CFloatVector, CFloatVector&> m_Transp; // transition probs
protected:
	CString m_sName;
};

class CHMMModelSet
{
public:
	CHMMModelSet();
	virtual ~CHMMModelSet(){};
	int Load(LPCTSTR sPath, DWORD dwFlags=0);
	int AddModel(){int n=m_Models.GetSize(); m_Models.SetSize(n + 1); return n;}

	int m_nChannels;
	CArray<CHMMModel, CHMMModel&> m_Models;

	CArray<CHMMState, CHMMState&> m_CommonStates;
protected:
	bool CopyState(CHMMModel* pCurHmm, int nCurState, char* sCurState);
	CStringArray m_CommonStateNames;
};

class CStateMeanIdxes
{
public:
	CStateMeanIdxes(){};
	~CStateMeanIdxes(){};

//	CArray<CWordArray, CWordArray&> ;
	CWordArray m_ChannelMeansIdxes; // nChannels * nMixes matrix
	int m_nMixes;
};

class CHMMLook  
{
public:
	CHMMLook();
	virtual ~CHMMLook();

	int LoadHMMs(LPCTSTR sPath, DWORD dwFlags=0);
	int DoAnalyseHmms();
	int AnalyseModel();

	CHMMModelSet* GetHMMSet(){return &m_HmmSet;};
	CHMMModel* GetCurModel(){
		return m_nCurModel>=0 && m_nCurModel<m_HmmSet.m_Models.GetSize() ? 
		&m_HmmSet.m_Models[m_nCurModel] : 0;
	};
	int GetCurModelIdx(){return m_nCurModel;};
	int SetCurModelIdx(int nModel)	{
		if(nModel>=0 && nModel<m_HmmSet.m_Models.GetSize()){m_nCurModel = nModel;} 
		VERIFY(AnalyseModel() == 0);
		return m_nCurModel;
	};
	float* GetProbValues(int nState, int& nValues);
	int GetChannels(){return m_HmmSet.m_nChannels;};

	int FillVectorsMinMax();
	int FillVectorsMinMaxGlobally(int nExceptChannel = -1);
	int UnifyMinMaxForChannels(int nStartChannel, int nChannels=-1);

	int GetChannelZeroIdx(int nState, int nChannel){
		if(nState < 0 || nChannel < 0 || nState>=m_arrChannelValueZeroIdxes.GetSize()){return -1;}
		if(nChannel >= m_arrChannelValueZeroIdxes[nState].GetSize()){return -1;}
		return m_arrChannelValueZeroIdxes[nState][nChannel];
	};
	int GetMeanIdx(int nState, int nChannel, int nMix){
		if(nState < 0 || nChannel < 0 || nMix < 0 || nState >= m_arrStateMeanIdxes.GetSize()){return -1;}
		CWordArray& arrMeans = m_arrStateMeanIdxes[nState].m_ChannelMeansIdxes;
		int nMixes = m_arrStateMeanIdxes[nState].m_nMixes;
		return nMix < nMixes ? arrMeans[nChannel * nMixes + nMix] : -1;
	};
	bool GetChannelMinMax(int nChan, float& min, float& max){
		if(nChan < 0 || nChan >= m_arrValueMinMax.GetSize()/2){return false;}
		min = m_arrValueMinMax[nChan * 2]; max = m_arrValueMinMax[nChan * 2 + 1];
		return true;
	};
	bool SetTestedValuesCount(int TotalTestValues){
		if(TotalTestValues <= 0){return false;}
		m_nTestedValues = TotalTestValues;
		return true;
	}

protected:
	CHMMModelSet m_HmmSet;
	CFloatVector m_arrValueMinMax; // min-maxes for each channel
	int m_nCurModel;
	int m_nTestedValues; // quantity of values tested against GMM
	CArray<CFloatVector, CFloatVector&> m_TestedStates; // States * m_nTestedValues matrix
	float GetValueProb(float x, float mean, float sigma);
	CArray<CWordArray, CWordArray&> m_arrChannelValueZeroIdxes; // [0:states*channels] index where channel value is closest to zero
	CArray<CStateMeanIdxes, CStateMeanIdxes&> m_arrStateMeanIdxes;

	CHMMLook* m_NextHMMLook; // next HMM look object in seria

};

#endif // !defined(AFX_HMMLOOK_H__D33E6AAB_01D6_463C_B928_93DBF2C31ACD__INCLUDED_)
